local module = {}

local ffi = require("ffi")

ffi.cdef [[
    typedef struct {
        double real;
        double imag;
    } lua_complex;
]]

local cos, sin, sqrt, atan, atan2, log, exp = math.cos, math.sin, math.sqrt, math.atan, math.atan2, math.log, math.exp

local complex_meta = {}
local ctype

local function complex_new(real, imag)
    local num = ctype()
    num.real = real
    num.imag = imag
    return num
end

function complex_meta.__add(a, b)
    return complex_new(a.real + b.real, a.imag + b.imag)
end

function complex_meta.__sub(a, b)
    return complex_new(a.real - b.real, a.imag - b.imag)
end

function complex_meta.__mul(a, b)
    if type(b) == "number" then
        return complex_new(a.real * b, a.imag * b)
    end

    return complex_new(a.real*b.real - a.imag*b.imag, a.real*b.imag + a.imag*b.real)
end

function complex_meta.__div(a, b)
    if type(b) == "number" then
        return complex_new(a.real / b, a.imag / b)
    end

    local val = b.real*b.real + b.imag*b.imag
    return complex_new((a.real * b.real + a.imag * b.imag) / val, (a.imag * b.real - a.real * b.imag) / val)
end

function complex_meta.__unm(a)
    return complex_new(-a.real, -a.imag)
end

function complex_meta.__pow(cx, scalar)
    if scalar % 1 == 0 then
        local real = cx.real
        local imag = cx.imag

        if scalar < 0 then
            local v = cx.real*cx.real + cx.imag*cx.imag
            real = real / v
            imag = imag / v
            scalar = -scalar
        end

        for i=2, scalar do
            real = real * cx.real - imag * cx.imag
            imag = real * cx.imag + imag * cx.real
        end

        return complex_new(real, imag)
    end

    local length = sqrt(cx.real*cx.real + cx.imag*cx.imag) ^ scalar
    local phi = atan2(cx.imag, cx.real) * scalar
    return complex_new(length * cos(phi), length * sin(phi))
end

function complex_meta.__eq(a, b)
    return a.real == b.real and a.imag == b.imag
end

function complex_meta.__tostring(cx)
    return string.format("%s + %si", tostring(cx.real), tostring(cx.imag))
end

function module.new(real, imag)
    if imag then
        return complex_new(real, imag)
    end

    return complex_new(real, 0)
end

function module.sqrt(cx)
    local len = sqrt(cx.real*cx.real + cx.imag*cx.imag)
    local sign = cx.imag < 0 and -1 or 1
    return complex_new(sqrt(cx.real + len)/2), sign * sqrt((len-cx.real)/2)
end

function module.ln(cx)
    return complex_new(log(sqrt(cx.real*cx.real + cx.imag*cx.imag)), atan2(cx.imag, cx.real))
end

function module.exp(cx)
    local expreal = exp(cx.real)
    return complex_new(expreal * cos(cx.imag), expreal * sin(cx.imag))
end

ctype = ffi.metatype("lua_complex", complex_meta)

return module