local complex = require("complex")
local bit = require("bit")

local band, bnot, bor = bit.band, bit.bnot, bit.bor

local function isPowerOfTwo(n)
    return band(n, n - 1) == 0
end

local function _fft(data, inverse)
    local dataLen = #data
    assert(isPowerOfTwo(dataLen), "data size must be a power of 2")
    
    -- rearrange data
    local target_index = 0
    local bit_mask

    for i=0, dataLen-1 do
        if target_index > i then
            local temp = data[target_index+1]
            data[target_index+1] = data[i+1]
            data[i+1] = temp
        end

        bit_mask = dataLen

        while true do
            --bit_mask = bit.rshift(bit_mask, 1)
            bit_mask = bit_mask / 2

            if (band(target_index, bit_mask)) == 0 then
                break
            end

            target_index = band(target_index, bnot(bit_mask))
        end

        target_index = bor(target_index, bit_mask)
    end

    local pi = inverse and math.pi or -math.pi

    local next, match
    local sine
    local delta
    local mult, factor, product

    do
        local i = 1

        while i < dataLen do
            next = i * 2
            delta = pi / i
            sine = math.sin(delta / 2)
            mult = complex.new(-2 * sine*sine, math.sin(delta))
            factor = complex.new(1.0)

            for j=0, i-1 do
                for k=j, dataLen-1, next do
                    match = k + i

                    product = data[match+1] * factor
                    data[match+1] = data[k+1] - product
                    data[k+1] = data[k+1] + product
                end

                factor = mult * factor + factor
            end

            i=i*2
        end
    end

    if inverse then
        for i, v in ipairs(data) do
            data[i] = v / dataLen
        end
    end
end

local function fft(data, dataLen, inverse)
    assert(isPowerOfTwo(dataLen), "data size must be a power of 2")
    
    -- rearrange data
    local target_index = 0
    local bit_mask

    for i=0, dataLen-1 do
        if target_index > i then
            local temp = data[target_index]
            data[target_index] = data[i]
            data[i] = temp
        end

        bit_mask = dataLen

        while true do
            --bit_mask = bit.rshift(bit_mask, 1)
            bit_mask = bit_mask / 2

            if (band(target_index, bit_mask)) == 0 then
                break
            end

            target_index = band(target_index, bnot(bit_mask))
        end

        target_index = bor(target_index, bit_mask)
    end

    local pi = inverse and math.pi or -math.pi

    local next, match
    local sine
    local delta
    local mult, factor, product

    do
        local i = 1

        while i < dataLen do
            next = i * 2
            delta = pi / i
            sine = math.sin(delta / 2)
            mult = complex.new(-2 * sine*sine, math.sin(delta))
            factor = complex.new(1.0)

            for j=0, i-1 do
                for k=j, dataLen-1, next do
                    match = k + i

                    product = data[match] * factor
                    data[match] = data[k] - product
                    data[k] = data[k] + product
                end

                factor = mult * factor + factor
            end

            i=i*2
        end
    end

    if inverse then
        for i=0, dataLen-1 do
            data[i] = data[i] / dataLen
        end
    end
end

local function ffi_fft(data, dataLen, inverse)
    assert(isPowerOfTwo(dataLen), "data size must be a power of 2")
    
    -- rearrange data
    local target_index = 0
    local bit_mask

    for i=0, dataLen-1 do
        if target_index > i then
            local temp = data[target_index]
            data[target_index] = data[i]
            data[i] = temp
        end

        bit_mask = dataLen

        while true do
            --bit_mask = bit.rshift(bit_mask, 1)
            bit_mask = bit_mask / 2

            if (band(target_index, bit_mask)) == 0 then
                break
            end

            target_index = band(target_index, bnot(bit_mask))
        end

        target_index = bor(target_index, bit_mask)
    end

    local pi = inverse and math.pi or -math.pi

    local next, match
    local sine
    local delta
    local mult, factor, product

    do
        local i = 1

        while i < dataLen do
            next = i * 2
            delta = pi / i
            sine = math.sin(delta / 2)
            mult = complex.new(-2 * sine*sine, math.sin(delta))
            factor = complex.new(1.0)

            for j=0, i-1 do
                for k=j, dataLen-1, next do
                    match = k + i

                    product = data[match] * factor
                    data[match] = data[k] - product
                    data[k] = data[k] + product
                end

                factor = mult * factor + factor
            end

            i=i*2
        end
    end

    if inverse then
        for i=0, dataLen-1 do
            data[i] = data[i] / dataLen
        end
    end
end

return {
    inplace = fft,
    real = function(data, inverse)
        local cx = {}

        for i, v in ipairs(data) do
            cx[i] = complex.new(v)
        end

        fft(cx, inverse)
        return cx
    end,

    ffi = {
        inplace = ffi_fft,
    }
}